// This include allows VTK to build on some platforms with broken Ruby
// header files.
#include "vtkRubyUtil.h"

#include "vtkSystemIncludes.h"

#include "vtkObject.h"
#include "vtkObjectFactory.h"
#include "vtkSmartPointerBase.h"
#include "vtkTimeStamp.h"
#include "vtkWindows.h"

#include <vtkstd/map>
#include <vtkstd/string>

#if defined ( _MSC_VER )
#  define vtkConvertPtrToLong(x) ((long)(PtrToUlong(x)))
#else
#  define vtkConvertPtrToLong(x) ((long)(x))
#endif


//#define VTKRUBYDEBUG

//--------------------------------------------------------------------
class vtkRubyDeleteCommand : public vtkCommand
{
public:
  static vtkRubyDeleteCommand *New(rbVTKObject *obj) {
    return new vtkRubyDeleteCommand(obj); };

  void Execute(vtkObject *caller, unsigned long, void *);

private:
  vtkRubyDeleteCommand(rbVTKObject *obj) : Self(obj) {};

  rbVTKObject *Self;
};

void vtkRubyDeleteCommand::Execute(vtkObject *caller,
                                     unsigned long vtkNotUsed(id),
                                     void *vtkNotUsed(data))
{
  rbVTKObject *obj = this->Self;
  if (obj->vtk_ptr != caller)
    {
    vtkGenericWarningMacro("Ruby vs. VTK mismatch for " << caller);
    return;
    }
  obj->vtk_ptr = NULL;
  obj->mark = Qnil;
}

//--------------------------------------------------------------------
static void rb_VTKObject_RuDelete(rbVTKObject *self)
{
  self->vtk_ptr->Delete();
  // the rest of the delection is handled when the VTK-level object
  // is destroyed
  //redef;
}

//--------------------------------------------------------------------
VALUE rb_VTKClass(const char* name)
{
  char *rb_class_name;
  VALUE klass;

  /*
  rb_class_name = ALLOC_N(char,strlen(name)+6);
  strcpy(rb_class_name, "Vtk::");
  strcat(rb_class_name, name);
  rb_class_name[5] = 'V';
  */
  name = name+3;
  rb_class_name = ALLOC_N(char,strlen(name)+6);
  strcpy(rb_class_name, "Vtk::");
  strcat(rb_class_name, name);
  klass = rb_eval_string(rb_class_name);
  free(rb_class_name);
  return klass;
}

//--------------------------------------------------------------------
VALUE rb_vtk_define_class_under(VALUE module, const char* name, VALUE super)
{
  /*
  char *rb_class_name;
  VALUE klass;

  rb_class_name = ALLOC_N(char,strlen(name)+1);
  strcpy(rb_class_name, name);
  rb_class_name[0] = 'V';
  klass = rb_define_class_under(module, rb_class_name, super);
  free(rb_class_name);
  return klass;
  */

  return rb_define_class_under(module, name+3, super);
}

//--------------------------------------------------------------------
#ifdef __cplusplus
extern "C" {
#endif
void rb_VTKObject_Mark(void *self)
{
  rbVTKObject *obj;
  int i;

  obj = (rbVTKObject *) self;
  if (rb_class_of(obj->mark)==rb_cArray)
    {
    rb_gc_mark(obj->mark);
    for (i=0;i<RARRAY(obj->mark)->len;i++)
      rb_gc_mark(RARRAY(obj->mark)->ptr[i]);
    }
}
void rb_VTKObject_Free(void *self)
{
  rbVTKObject *obj = (rbVTKObject *) self;
  vtkRubyUnRegister(obj);
  free(obj);
}
#ifdef __cplusplus
}
#endif


VALUE rb_VTKObject_New(VALUE klass, vtkObjectBase *ptr)
{
  rbVTKObject *obj;

  if (!ptr)
    rb_raise(rb_eArgError,
                    "this is an abstract class and cannot be instantiated");
  ptr->Register(NULL);
  ptr->Delete();
  obj = ALLOC(rbVTKObject);
  obj->vtk_ptr = ptr;
  obj->ref_count = 1;
  obj->mark = rb_ary_new();
  return Data_Wrap_Struct(klass,rb_VTKObject_Mark,rb_VTKObject_Free,obj);
}

void rb_VTKObject_PushMark(VALUE self, VALUE other)
{
  rbVTKObject *obj;
  if (rb_obj_is_kind_of(self,rb_vtkObjectBaseClass()))
    {
    Data_Get_Struct(self,rbVTKObject,obj);
    if (rb_class_of(obj->mark)==rb_cArray)
      {
      rb_ary_push(obj->mark, other);
      }
    }
}

VALUE rb_VTKSpecialObject_New(VALUE klass, void *ptr)
{
  rbVTKSpecialObject *obj;

  obj = ALLOC(rbVTKSpecialObject);
  obj->ptr = ptr;
  obj->ref_count = 1;
  obj->mark = rb_ary_new();

  return Data_Wrap_Struct(klass,rb_VTKObject_Mark,rb_VTKObject_Free,obj);
}

//--------------------------------------------------------------------
VALUE vtkRubyGetObjectFromPointer(vtkObjectBase *ptr)
{
  return vtkRubyGetObjectFromPointer(ptr, rb_VTKClass(ptr->GetClassName()));
}

VALUE vtkRubyGetObjectFromPointer(vtkObjectBase *ptr, VALUE klass)
{
  rbVTKObject *obj;

  if (ptr)
    ptr->Register(NULL);
  else
    return Qnil;

  obj = ALLOC(rbVTKObject);
  obj->vtk_ptr = ptr;
  obj->ref_count = 1;
  obj->mark = rb_ary_new();

  return Data_Wrap_Struct(klass,rb_VTKObject_Mark,rb_VTKObject_Free,obj);

}

//--------------------------------------------------------------------
vtkObjectBase *vtkRubyGetPointerFromObject(VALUE robj)
{ 
  rbVTKObject *obj;
  vtkObjectBase *ptr;

  // convert Qnil to NULL every time
  if (robj == Qnil)
    {
      return NULL;
    }

  // check to ensure it is a vtk object
  if (rb_obj_is_kind_of(robj,rb_vtkObjectBaseClass()))
    {
      Data_Get_Struct(robj,rbVTKObject,obj);
      ptr = obj->vtk_ptr;
    }
  else
   {
#ifdef VTKRUBYDEBUG
    vtkGenericWarningMacro("Object " << robj << " is not a VTK object!!");
#endif  
    rb_raise(rb_eRuntimeError,"method requires a VTK object");
    }
  
#ifdef VTKRUBYDEBUG
  vtkGenericWarningMacro("Checking into obj " << obj << " ptr = " << ptr);
#endif  

#ifdef VTKRUBYDEBUG
    vtkGenericWarningMacro("Got obj= " << obj << " ptr= " << ptr << " " << result_type);
#endif  
    return ptr;
}

//--------------------------------------------------------------------
VALUE vtkRubyGetObjectFromObject(VALUE arg, const char *type, VALUE klass)
{
  char *ptrText = STR2CSTR(arg);

  vtkObjectBase *ptr;
  char typeCheck[256];  // typeCheck is currently not used
  int i = sscanf(ptrText,"_%lx_%s",(long *)&ptr,typeCheck);

  if (i <= 0)
    {
    i = sscanf(ptrText,"Addr=0x%lx",(long *)&ptr);
    }
  if (i <= 0)
    {
    i = sscanf(ptrText,"%lx",(long *)&ptr);
    }
  if (i <= 0)
    {
    rb_raise(rb_eArgError,"could not extract hexidecimal address from argument string");
    return Qnil;
    }

  if (!ptr->IsA(type))
    {
    char error_string[256];
    sprintf(error_string,"method requires a %s address, a %s address was provided.",
          type,((vtkObjectBase *)ptr)->GetClassName());
    rb_raise(rb_eArgError,error_string);
    return Qnil;
    }

  return vtkRubyGetObjectFromPointer(ptr, klass);
}

//--------------------------------------------------------------------
void vtkRubyUnRegister(VALUE self)
{
  rbVTKObject *obj;
  if (rb_obj_is_kind_of(self,rb_vtkObjectBaseClass()))
    {
    Data_Get_Struct(self,rbVTKObject,obj);
    vtkRubyUnRegister(obj);
    }
}
void vtkRubyUnRegister(rbVTKObject *obj)
{
  if (obj->ref_count)
    {
    obj->vtk_ptr->UnRegister(NULL);
    obj->ref_count = 0;
    }
}

//--------------------------------------------------------------------
// mangle a void pointer into a  SWIG-type string
char *vtkRubyManglePointer(void *ptr, const char *type)
{
  static char ptrText[128];
  sprintf(ptrText,"_%*.*lx_%s",2*(int)sizeof(void *),2*(int)sizeof(void *),
	  vtkConvertPtrToLong(ptr),type);
  return ptrText;
}

//--------------------------------------------------------------------
// unmangle a void pointer from a SWIG-style string
void *vtkRubyUnmanglePointer(char *ptrText, int *len, const char *type)
{
  int i;
  void *ptr;
  char typeCheck[128];
  if (*len < 128)
    {
    i = sscanf(ptrText,"_%lx_%s",(long *)&ptr,typeCheck);
    if (strcmp(type,typeCheck) == 0)
      { // sucessfully unmangle
      *len = 0;
      return ptr;
      }
    else if (i == 2)
      { // mangled pointer of wrong type
      *len = -1;
      return NULL;
      }
    }
  // couldn't unmangle: return string as void pointer if it didn't look
  // like a SWIG mangled pointer
  return (void *)ptrText;
}

//--------------------------------------------------------------------
// These functions check an array that was sent to a method to see if
// any of the values were changed by the method.
// If a value was changed, then the corresponding value in the ruby
// list is modified.

template<class T>
static inline
int vtkRubyCheckFloatArray(VALUE *args, int i, T *a, int n)
{
  int changed = 0;

  VALUE seq = args[i];
  for (i = 0; i < n; i++)
    {
    VALUE oldobj = rb_ary_entry(seq, i);
    T oldval = (T)NUM2DBL(oldobj);
    changed |= (a[i] != oldval);
    }

  if (changed)
    {
    for (i = 0; i < n; i++)
      {
      VALUE newobj = rb_float_new(a[i]);
      rb_ary_store(seq, i, newobj);
      }
    }

  return 0;
}

template<class T>
static inline
int vtkRubyCheckIntArray(VALUE *args, int i, T *a, int n)
{
  int changed = 0;

  VALUE seq = args[i];
  for (i = 0; i < n; i++)
    {
    VALUE oldobj = rb_ary_entry(seq, i);
    T oldval = (T)NUM2LONG(oldobj);
    changed |= (a[i] != oldval);
    }

  if (changed)
    {
    for (i = 0; i < n; i++)
      {
      VALUE newobj = LONG2NUM(a[i]);
      rb_ary_store(seq, i, newobj);
      }
    }

  return 0;
}

#if defined(VTK_TYPE_USE_LONG_LONG) || defined(VTK_TYPE_USE___INT64)
template<class T>
static inline
int vtkRubyCheckLongArray(VALUE *args, int i, T *a, int n)
{
  int changed = 0;

  VALUE seq = args[i];
  for (i = 0; i < n; i++)
    {
    VALUE oldobj = rb_ary_entry(seq, i);
    T oldval;
#ifdef PY_LONG_LONG
    oldval = NUM2LL(oldobj);
#else
    oldval = NUM2LONG(oldobj);
#endif
    changed |= (a[i] != oldval);
    }

  if (changed)
    {
    for (i = 0; i < n; i++)
      {
#if defined(VTK_TYPE_USE_LONG_LONG)
# if defined(PY_LONG_LONG) && (VTK_SIZEOF_LONG != VTK_SIZEOF_LONG_LONG)
      VALUE newobj = LL2NUM(a[i]);
# else
      VALUE newobj = LONG2NUM((long)a[i]);
# endif
#else
# if defined(PY_LONG_LONG) && (VTK_SIZEOF_LONG != VTK_SIZEOF___INT64)
      VALUE newobj = LL2NUM(a[i]);
# else
      VALUE newobj = LONG2NUM((long)a[i]);
# endif
#endif
      rb_ary_store(seq, i, newobj);
      }
    }

  return 0;
}
#endif

int vtkRubyCheckArray(VALUE *args, int i, char *a, int n)
{
  return vtkRubyCheckIntArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, signed char *a, int n)
{
  return vtkRubyCheckIntArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, unsigned char *a, int n)
{
  return vtkRubyCheckIntArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, short *a, int n)
{
  return vtkRubyCheckIntArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, unsigned short *a, int n)
{
  return vtkRubyCheckIntArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, int *a, int n)
{
  return vtkRubyCheckIntArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, unsigned int *a, int n)
{
  return vtkRubyCheckIntArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, long *a, int n)
{
  return vtkRubyCheckIntArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, unsigned long *a, int n)
{
  return vtkRubyCheckIntArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, float *a, int n)
{
  return vtkRubyCheckFloatArray(args, i, a, n);
}

int vtkRubyCheckArray(VALUE *args, int i, double *a, int n)
{
  return vtkRubyCheckFloatArray(args, i, a, n);
}

#if defined(VTK_TYPE_USE_LONG_LONG)
int vtkRubyCheckArray(VALUE *args, int i, long long *a, int n)
{
  return vtkRubyCheckLongArray(args, i, a, n);
}
int vtkRubyCheckArray(VALUE *args, int i, unsigned long long *a, int n)
{
  return vtkRubyCheckLongArray(args, i, a, n);
}
#endif

#if defined(VTK_TYPE_USE___INT64)
int vtkRubyCheckArray(VALUE *args, int i, __int64 *a, int n)
{
  return vtkRubyCheckLongArray(args, i, a, n);
}
int vtkRubyCheckArray(VALUE *args, int i, unsigned __int64 *a, int n)
{
  return vtkRubyCheckLongArray(args, i, a, n);
}
#endif


//--------------------------------------------------------------------
void vtkRubyVoidFunc(void *arg)
{
  VALUE result;
  VALUE func = (VALUE)arg;

  result = rb_funcall(func,rb_intern("call"),0);

}

//--------------------------------------------------------------------
void vtkRubyVoidFuncArgDelete(void *arg)
{
}


//--------------------------------------------------------------------
vtkRubyCommand::vtkRubyCommand()
{ 
  this->obj = Qnil;
}

vtkRubyCommand::~vtkRubyCommand()
{ 
  this->obj = Qnil;
}

void vtkRubyCommand::SetObject(VALUE o)
{ 
  this->obj = o; 
}

void vtkRubyCommand::Execute(vtkObject *ptr, unsigned long eventtype,
                               void *CallData)
{
  VALUE obj2;
  const char *eventname;
  VALUE argv[3];

  if (ptr && ptr->GetReferenceCount() > 0)
    {
    obj2 = vtkRubyGetObjectFromPointer(ptr);
    rb_funcall(obj2, rb_intern("UnRegister"), 1, Qnil);
    }
  else
    {
    obj2 = Qnil;
    }

    eventname = this->GetStringFromEventId(eventtype);


    argv[0] = obj2;
    argv[1] = rb_str_new2(eventname);
    if (CallData==NULL){
      argv[2] = Qnil;}
    else
      argv[2] = rb_str_new2((char*)CallData);
    rb_funcall2(this->obj, rb_intern("call"), 3, argv);
}
//--------------------------------------------------------------------



