/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkWrapRuby.c,v $

  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/

#include "vtkRuby.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "vtkParse.h"
#include "vtkConfigure.h"

int numberOfWrappedFunctions = 0;
FunctionInfo *wrappedFunctions[1000];
extern FunctionInfo *currentFunction;

static int class_has_new = 0;

/* when the cpp file doesn't have enough info use the hint file */
void use_hints(FILE *fp)
{
  int  i;
  
  switch (currentFunction->ReturnType % 0x1000)
    {
    case 0x301:
    case 0x307:
      fprintf(fp,"    return rb_ary_new3(%d", currentFunction->HintSize);
      for (i = 0; i < currentFunction->HintSize; i++)
        {
        fprintf(fp,",rb_float_new(temp%i[%d])",MAX_ARGS,i);
        }
      fprintf(fp,");\n");
      break;
    case 0x304:
    case 0x30D:
      fprintf(fp,"    return rb_ary_new3(%d", currentFunction->HintSize);
      for (i = 0; i < currentFunction->HintSize; i++) 
        {
        fprintf(fp,",INT2NUM(temp%i[%d])",MAX_ARGS,i);
        }
      fprintf(fp,");\n");
      break;
    case 0x30A:
      fprintf(fp,"    return rb_ary_new3(%d", currentFunction->HintSize);
      for (i = 0; i < currentFunction->HintSize; i++) 
        {
#ifdef VTK_USE_64BIT_IDS
#ifdef HAVE_LONG_LONG
        fprintf(fp,",LL2NUM(temp%i[%d])",MAX_ARGS,i);
#else
        fprintf(fp,",LONG2NUM(temp%i[%d])",MAX_ARGS,i);
#endif
#else
        fprintf(fp,",INT2NUM(temp%i[%d])",MAX_ARGS,i);
#endif
        }
      fprintf(fp,");\n");
      break;
    case 0x30B: case 0x30C:
      fprintf(fp,"    return rb_ary_new3(%d", currentFunction->HintSize);
      for (i = 0; i < currentFunction->HintSize; i++) 
        {
#ifdef HAVE_LONG_LONG
        fprintf(fp,",LL2NUM(temp%i[%d])",MAX_ARGS,i);
#else
        fprintf(fp,",LONG2NUM(temp%i[%d])",MAX_ARGS,i);
#endif
        }
      fprintf(fp,");\n");
      break;
    case 0x305: case 0x306: case 0x313: case 0x314:
    case 0x31A: case 0x31B: case 0x31C: case 0x315: case 0x316:
      break;
    }
  return;
}


void output_temp(FILE *fp, int i, int aType, char *Id, int aCount)
{
  /* handle VAR FUNCTIONS */
  if (aType == 0x5000)
    {
    fprintf(fp,"    VALUE temp%i;\n",i); 
    return;
    }
  
  if (((aType % 0x10) == 0x2)&&(!((aType % 0x1000)/0x100)))
    {
    return;
    }

  /* for const * return types prototype with const */
  if ((i == MAX_ARGS)&&(aType % 0x2000 >= 0x1000))
    {
    fprintf(fp,"    const ");
    }
  else
    {
    fprintf(fp,"    ");
    }
  
  if ((aType % 0x100)/0x10 == 0x1)
    {
    fprintf(fp,"unsigned ");
    }
  
  switch (aType % 0x10)
    {
    case 0x1:   fprintf(fp,"float  "); break;
    case 0x7:   fprintf(fp,"double "); break;
    case 0x4:   fprintf(fp,"int    "); break;
    case 0x5:   fprintf(fp,"short  "); break;
    case 0x6:   fprintf(fp,"long   "); break;
    case 0x2:   fprintf(fp,"void   "); break;
    case 0x3:   fprintf(fp,"char   "); break;
    case 0x9:   fprintf(fp,"%s ",Id); break;
    case 0xA:   fprintf(fp,"vtkIdType "); break;
    case 0xB:   fprintf(fp,"long long "); break;
    case 0xC:   fprintf(fp,"__int64 "); break;
    case 0xD:   fprintf(fp,"signed char "); break;
    case 0x8: return;
    }
  
  switch ((aType % 0x1000)/0x100)
    {
    case 0x1: fprintf(fp, " *"); break; /* act " &" */
    case 0x2: fprintf(fp, "&&"); break;
    case 0x3: 
      if ((i == MAX_ARGS)||(aType % 0x10 == 0x9)||(aType % 0x1000 == 0x303)
          ||(aType % 0x1000 == 0x302))
        {
        fprintf(fp, " *"); 
        }
      break;
    case 0x4: fprintf(fp, "&*"); break;
    case 0x5: fprintf(fp, "*&"); break;
    case 0x7: fprintf(fp, "**"); break;
    default: fprintf(fp,"  "); break;
    }
  fprintf(fp,"temp%i",i);
  
  /* handle arrays */
  if ((aType % 0x1000/0x100 == 0x3)&&
      (i != MAX_ARGS)&&(aType % 0x10 != 0x9)&&(aType % 0x1000 != 0x303)
      &&(aType % 0x1000 != 0x302))
    {
    fprintf(fp,"[%i]",aCount);
    }

  fprintf(fp,";\n");
  if (aType % 0x1000 == 0x302 && i != MAX_ARGS)
    {
    fprintf(fp,"    int      size%d;\n",i);
    }
}

void do_return(FILE *fp)
{
  /* ignore void */
  if (((currentFunction->ReturnType % 0x10) == 0x2)&&
      (!((currentFunction->ReturnType % 0x1000)/0x100)))
    {
    fprintf(fp,"    return Qnil;\n");
    return;
    }
  
  switch (currentFunction->ReturnType % 0x1000)
    {
    case 0x303:
      fprintf(fp,"    if (temp%i == NULL) {\n",MAX_ARGS);
      fprintf(fp,"      return Qnil;\n      }\n");
      fprintf(fp,"    else {\n");
      fprintf(fp,"      return rb_str_new2(temp%i);\n      }\n",MAX_ARGS);
    break;
    case 0x109:
    case 0x309:  
      {
      fprintf(fp,"    {\n");
      fprintf(fp,"    return vtkRubyGetObjectFromPointer((vtkObjectBase *)temp%i);\n",
              MAX_ARGS, currentFunction->ReturnClass, currentFunction->ReturnClass);
      fprintf(fp,"    }\n");
      break;
      }
      
    /* handle functions returning vectors */
    /* this is done by looking them up in a hint file */
    case 0x301: case 0x307: case 0x30A: case 0x30B: case 0x30C: case 0x30D:
    case 0x304: case 0x305: case 0x306:
      use_hints(fp);
      break;
    case 0x302:
      {
      fprintf(fp,"    if (temp%i == NULL)\n        {\n",MAX_ARGS);
      fprintf(fp,"      return Qnil;\n        }\n");
      fprintf(fp,"    else\n        {\n");
      fprintf(fp,"      return rb_str_new2(vtkRubyManglePointer(temp%i,\"void_p\"));\n        }\n",
              MAX_ARGS);
      break;
      }
    case 0x1:
    case 0x7:
      {
      fprintf(fp,"    return rb_float_new(temp%i);\n",
                      MAX_ARGS);
      break;
      }
    case 0x13:
    case 0x14:
    case 0x15:
    case 0x4:
    case 0x5:
    case 0x6:
    case 0xD:
      {
      fprintf(fp,"    return INT2NUM(temp%i);\n", MAX_ARGS); 
      break;
      }
    case 0x16:   
      {
      fprintf(fp,"    return UINT2NUM(temp%i);\n", MAX_ARGS);
      break;
      }
#if defined(VTK_USE_64BIT_IDS) && defined(HAVE_LONG_LONG) && (VTK_SIZEOF_LONG != VTK_SIZEOF_ID_TYPE)
    case 0xA:
      {
      fprintf(fp,"    return LL2NUM(temp%i);\n", MAX_ARGS);
      break;
      }
    case 0x1A:
      {
      fprintf(fp,"    return ULL2NUM(temp%i);\n",
              MAX_ARGS);
      break;
      }
#else
    case 0xA:
      {
      fprintf(fp,"    return LONG2NUM((long)temp%i);\n", MAX_ARGS);
      break;
      }
    case 0x1A:
      {
#if (PY_VERSION_HEX >= 0x02020000)
      fprintf(fp,"    return ULONG2NUM((unsigned long)temp%i);\n",
              MAX_ARGS);
#else
      fprintf(fp,"    return LONG2NUM((long)temp%i);\n",
              MAX_ARGS);
#endif
      break;
      }
#endif
#if defined(VTK_SIZEOF_LONG_LONG)
# if defined(HAVE_LONG_LONG) && (VTK_SIZEOF_LONG != VTK_SIZEOF_LONG_LONG)
    case 0xB:
      {
      fprintf(fp,"    return LL2NUM(temp%i);\n", MAX_ARGS);
      break;
      }
    case 0x1B:
      {
      fprintf(fp,"    return ULL2NUM(temp%i);\n",
              MAX_ARGS);
      break;
      }
# else
    case 0xB:
      {
      fprintf(fp,"    return LONG2NUM(temp%i);\n", MAX_ARGS);
      break;
      }
    case 0x1B:
      {
      fprintf(fp,"    return ULONG2NUM(temp%i);\n",
              MAX_ARGS);
      break;
      }
# endif
#endif
#if defined(VTK_SIZEOF___INT64)
# if defined(HAVE_LONG_LONG) && (VTK_SIZEOF_LONG != VTK_SIZEOF___INT64)
    case 0xC:
      {
      fprintf(fp,"    return LL2NUM(temp%i);\n", MAX_ARGS);
      break;
      }
    case 0x1C:
      {
      fprintf(fp,"    return ULL2NUM(temp%i);\n",
              MAX_ARGS);
      break;
      }
# else
    case 0xC:
      {
      fprintf(fp,"    return LONG2NUM(temp%i);\n", MAX_ARGS);
      break;
      }
    case 0x1C:
      {
      fprintf(fp,"    return ULONG2NUM(temp%i);\n",
              MAX_ARGS);
      break;
      }
# endif
#endif
    case 0x3:   
      {
      fprintf(fp,"    return rb_str_new((char *)&temp%i,1);\n",
              MAX_ARGS);
      break;
      }
    }
}

void get_args(FILE *fp,int occ)
{
  int argtype;
  int i, j;

  if (currentFunction->ArgTypes[0] == 0x5000)
    {
    fprintf(fp,"    if (rb_class_of(argv[0])!=rb_cProc)\n");
    fprintf(fp,"      goto break%d;\n",occ);
    fprintf(fp,"    temp0 = argv[0];\n");
    fprintf(fp,"    rb_VTKObject_PushMark(self, temp0);\n");
    return;
    }

  for (i = 0; i < currentFunction->NumberOfArguments; i++)
    {
    argtype = currentFunction->ArgTypes[i] % 0x1000;

    switch (argtype)
      {
      case 0x301:
        fprintf(fp,"    if (rb_class_of(argv[%d])!=rb_cArray||RARRAY(argv[%d])->len!=%d)\n",i,i,currentFunction->ArgCounts[i]);
        fprintf(fp,"      goto break%d;\n",occ);
        for (j = 0; j < currentFunction->ArgCounts[i]; j++) 
          {
          fprintf(fp,"    temp%i[%d] = (float)NUM2DBL(rb_ary_entry(argv[%d],%d));\n",i,j,i,j);
          }
        break;
      case 0x307:  
        fprintf(fp,"    if (rb_class_of(argv[%d])!=rb_cArray||RARRAY(argv[%d])->len!=%d)\n",i,i,currentFunction->ArgCounts[i]);
        fprintf(fp,"      goto break%d;\n",occ);
        for (j = 0; j < currentFunction->ArgCounts[i]; j++) 
          {
          fprintf(fp,"    temp%i[%d] = NUM2DBL(rb_ary_entry(argv[%d],%d));\n",i,j,i,j);
          }
        break;
      case 0x304:
        fprintf(fp,"    if (rb_class_of(argv[%d])!=rb_cArray||RARRAY(argv[%d])->len!=%d)\n",i,i,currentFunction->ArgCounts[i]);
        fprintf(fp,"      goto break%d;\n",occ);
        for (j = 0; j < currentFunction->ArgCounts[i]; j++) 
          {
          fprintf(fp,"    temp%i[%d] = NUM2INT(rb_ary_entry(argv[%d],%d));\n",i,j,i,j);
          }
        break;
      case 0x30A:
        fprintf(fp,"    if (rb_class_of(argv[%d])!=rb_cArray||RARRAY(argv[%d])->len!=%d)\n",i,i,currentFunction->ArgCounts[i]);
        fprintf(fp,"      goto break%d;\n",occ);
        for (j = 0; j < currentFunction->ArgCounts[i]; j++) 
          {
#ifdef VTK_USE_64BIT_IDS
#ifdef HAVE_LONG_LONG
          fprintf(fp,"    temp%i[%d] = NUM2LL(rb_ary_entry(argv[%d],%d));\n",i,j,i,j);
#else
          fprintf(fp,"    temp%i[%d] = NUM2LONG(rb_ary_entry(argv[%d],%d));\n",i,j,i,j);
#endif
#else
          fprintf(fp,"    temp%i[%d] = NUM2INT(rb_ary_entry(argv[%d],%d));\n",i,j,i,j);
#endif
          }
        break;
      case 0x30B: case 0x30C:
        fprintf(fp,"    if (rb_class_of(argv[%d])!=rb_cArray||RARRAY(argv[%d])->len!=%d)\n",i,i,currentFunction->ArgCounts[i]);
        fprintf(fp,"      goto break%d;\n",occ);
        for (j = 0; j < currentFunction->ArgCounts[i]; j++) 
          {
#ifdef HAVE_LONG_LONG
          fprintf(fp,"    temp%i[%d] = NUM2LL(rb_ary_entry(argv[%d],%d));\n",i,j,i,j);
#else
          fprintf(fp,"    temp%i[%d] = NUM2LONG(rb_ary_entry(argv[%d],%d));\n",i,j,i,j);
#endif
          }
        break;
      case 0x109:
      case 0x309:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_vtkObjectBaseClass())&&argv[%d]!=Qnil)\n",i,i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = (%s *)vtkRubyGetPointerFromObject(argv[%d]);\n",i,currentFunction->ArgClasses[i],i);
        fprintf(fp,"    if (temp%i!=NULL&&!((vtkObjectBase *)temp%i)->IsA(\"%s\"))\n",i,i,currentFunction->ArgClasses[i]);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    rb_VTKObject_PushMark(self, argv[%d]);\n",i);
	break;
      case 0x303:
        fprintf(fp,"    if (argv[%d]!=Qnil&&rb_class_of(argv[%d])!=rb_cString)\n",i,i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = argv[%d]==Qnil?NULL:STR2CSTR(argv[%d]);\n",i,i,i);
	fprintf(fp,"    rb_VTKObject_PushMark(self, argv[%d]);\n",i);
	break;
      case 0x302:
        fprintf(fp,"    if (argv[%d]!=Qnil&&rb_class_of(argv[%d])!=rb_cString)\n",i,i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    if (argv[%d]==Qnil)\n",i);
	fprintf(fp,"      {\n");
	fprintf(fp,"        temp%i = (void*)STR2CSTR(argv[%d]);\n",i,i);
	fprintf(fp,"        size%i = RSTRING(argv[%d])->len;\n",i,i);
	fprintf(fp,"      }\n");
	fprintf(fp,"    else\n");
	fprintf(fp,"      {\n");
	fprintf(fp,"        temp%i = NULL;\n",i);
	fprintf(fp,"        size%i = 0;\n",i);
	fprintf(fp,"      }\n");
	fprintf(fp,"    rb_VTKObject_PushMark(self, argv[%d]);\n",i);
	break;
      case 0x1:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = (float)NUM2DBL(argv[%d]);\n",i,i);
	break;
      case 0x7:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = NUM2DBL(argv[%d]);\n",i,i);
	break;
      case 0x14:
      case 0x4:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = NUM2INT(argv[%d]);\n",i,i);
	break;
      case 0x15:
      case 0x5:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = (short)NUM2INT(argv[%d]);\n",i,i);
	break;
      case 0x16:
      case 0x6:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = NUM2LONG(argv[%d]);\n",i,i);
	break;
      case 0x1A:
      case 0xA:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
#ifdef VTK_USE_64BIT_IDS
#ifdef HAVE_LONG_LONG
	fprintf(fp,"    temp%i = NUM2LL(argv[%d]);\n",i,i);
#else
	fprintf(fp,"    temp%i = NUM2LONG(argv[%d]);\n",i,i);
#endif
#else
	fprintf(fp,"    temp%i = NUM2INT(argv[%d]);\n",i,i);
#endif
	break;
      case 0xB: case 0xC:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
#ifdef HAVE_LONG_LONG
	fprintf(fp,"    temp%i = NUM2LL(argv[%d]);\n",i,i);
#else
	fprintf(fp,"    temp%i = NUM2LONG(argv[%d]);\n",i,i);
#endif
	break;
      case 0x1B: case 0x1C:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
#ifdef HAVE_LONG_LONG
	fprintf(fp,"    temp%i = NUM2ULL(argv[%d]);\n",i,i);
#else
	fprintf(fp,"    temp%i = NUM2ULONG(argv[%d]);\n",i,i);
#endif
	break;
      case 0xD:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = (signed char)NUM2INT(argv[%d]);\n",i,i);
	break;
      case 0x3:
        fprintf(fp,"    if (rb_class_of(argv[%d])!=rb_cString)\n",i,i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = STR2CSTR(argv[%d])[0];\n",i,i);
	break;
      case 0x13:
        fprintf(fp,"    if (!rb_obj_is_kind_of(argv[%d],rb_cNumeric))\n",i);
        fprintf(fp,"      goto break%d;\n",occ);
	fprintf(fp,"    temp%i = (char)NUM2INT(argv[%d]);\n",i,i);
	break;
      }
    }
  return;
}


void outputFunction2(FILE *fp, FileInfo *data)
{
  int i, k, is_static, is_vtkobject, fnum, occ, backnum;
  FunctionInfo *theFunc;
  FunctionInfo *backFunc;

  is_vtkobject = ((strcmp(data->ClassName,"vtkObjectBase") == 0) || 
                  (data->NumberOfSuperClasses != 0));

  /* create a signature for each method (for use in docstring) */
  for (fnum = 0; fnum < numberOfWrappedFunctions; fnum++)
    {
    theFunc = wrappedFunctions[fnum];
    currentFunction = theFunc;
    }

  /* create external type declarations for all object
     return types */
  for (fnum = 0; fnum < numberOfWrappedFunctions; fnum++)
    {
    theFunc = wrappedFunctions[fnum];
    currentFunction = theFunc;

    /* check for object return types */
    if ((theFunc->ReturnType % 0x1000 == 0x309)||
        (theFunc->ReturnType % 0x1000 == 0x109))
      {
      /* check that we haven't done this type (no duplicate declarations) */
      for (backnum = fnum-1; backnum >= 0; backnum--) 
        {
        backFunc = wrappedFunctions[backnum];
        if (((backFunc->ReturnType % 0x1000 == 0x309)||
             (backFunc->ReturnType % 0x1000 == 0x109)) &&
            (strcmp(theFunc->ReturnClass,backFunc->ReturnClass) == 0))
          {
          break;
          }
        }
      }
    }

  /* for each function in the array */
  for (fnum = 0; fnum < numberOfWrappedFunctions; fnum++)
    {
    /* make sure we haven't already done one of these */
    theFunc = wrappedFunctions[fnum];
    currentFunction = theFunc;

    if (theFunc->Name)
      {
      fprintf(fp,"\n");

      /* check whether all signatures are static methods */
      is_static = 1;
      for (occ = fnum; occ < numberOfWrappedFunctions; occ++)
        {
        /* is it the same name */
        if (wrappedFunctions[occ]->Name &&
            !strcmp(theFunc->Name,wrappedFunctions[occ]->Name))
          {
          /* check for static methods */
          if (((wrappedFunctions[occ]->ReturnType/0x1000) & 0x2) != 0x2)
            {
            is_static = 0;
            }
          }
        }
        
      if(currentFunction->IsLegacy)
        {
        fprintf(fp,"#if !defined(VTK_LEGACY_REMOVE)\n");
        }
      fprintf(fp,"static VALUE rb_%s_%s(int argc, VALUE *argv, VALUE self)\n",
              data->ClassName,currentFunction->Name);
      fprintf(fp,"{\n");

      if (strcmp(currentFunction->Name,"UnRegister")==0)
	{
        fprintf(fp,"  if (argc == 1)\n");
        fprintf(fp,"    {\n");
        for (i = 0; i < currentFunction->NumberOfArguments; i++)
          {
          output_temp(fp, i, currentFunction->ArgTypes[i],
                      currentFunction->ArgClasses[i],
                      currentFunction->ArgCounts[i]);
          }
        output_temp(fp, MAX_ARGS,currentFunction->ReturnType,
                    currentFunction->ReturnClass,0);
        get_args(fp,fnum);
	fprintf(fp,"    vtkRubyUnRegister(self);\n");
	fprintf(fp,"    return Qnil;\n");
        fprintf(fp,"    }\n");
        fprintf(fp,"  break%d:\n",fnum);
	}
      else
	{
        /* find all occurances of this method */
        for (occ = fnum; occ < numberOfWrappedFunctions; occ++)
	  {
          is_static = 0;

	  /* is it the same name */
	  if (wrappedFunctions[occ]->Name && 
	      !strcmp(theFunc->Name,wrappedFunctions[occ]->Name))
	    {
             /* check for static methods */
            if (((wrappedFunctions[occ]->ReturnType/0x1000) & 0x2) == 0x2)
	      {
              is_static = 1;
	      }

	    currentFunction = wrappedFunctions[occ];

	    fprintf(fp,"  /* handle an occurrence */\n");
	    fprintf(fp,"  if (argc == %d)\n", currentFunction->NumberOfArguments);
	    fprintf(fp,"    {\n");
	    /* declare the variables */
	    if (!is_static)
	      {
	      if (is_vtkobject)
		{
		fprintf(fp,"    %s *op;\n\n",data->ClassName);
		}
	      else 
		{
		fprintf(fp,"    %s *op = (%s *)vtkRubyGetPointerFromObject(self)->ptr;\n\n",data->ClassName,data->ClassName);
		}
	      }

	    /* process the args */
	    for (i = 0; i < currentFunction->NumberOfArguments; i++)
	      {
	      output_temp(fp, i, currentFunction->ArgTypes[i],
			  currentFunction->ArgClasses[i],
			  currentFunction->ArgCounts[i]);
	      }
	    output_temp(fp, MAX_ARGS,currentFunction->ReturnType,
			currentFunction->ReturnClass,0);
	    if (is_static || !is_vtkobject)
	      {
	      get_args(fp,occ);
	      }
	    else
	      {
	      get_args(fp,occ);
	      fprintf(fp,"    op = (%s *)vtkRubyGetPointerFromObject(self);\n",data->ClassName);
	      }

	    /* make sure passed method is callable  for VAR functions */
	    if (currentFunction->NumberOfArguments == 0x1 &&
		currentFunction->ArgTypes[0] == 0x5000)
	      {
	      fprintf(fp,"    if (rb_class_of(temp0)!=rb_cProc)\n");
	      fprintf(fp,"      {\n      rb_raise(rb_eArgError,\"vtk callback method passed to %s in %s was not callable.\");\n",
		      currentFunction->Name,data->ClassName);
	      fprintf(fp,"      }\n");
	      }

	    /* check for void pointers and pass appropriate info*/
	    for (i = 0; i < currentFunction->NumberOfArguments; i++)
	      {
	      if (currentFunction->ArgTypes[i] % 0x1000 == 0x302)
		{
		fprintf(fp,"    temp%i = vtkRubyUnmanglePointer((char *)temp%i,&size%i,(char*)\"%s\");\n",i,i,i,"void_p");
		fprintf(fp,"    if (size%i == -1) {\n      rb_raise(rb_eRuntimeError,\"mangled pointer to %s in %s was of incorrect type.\");\n",
			i,currentFunction->Name,data->ClassName);
		fprintf(fp,"      }\n");
		fprintf(fp,"    else if (size%i == -2) {\n      rb_raise(rb_eRuntimeError,\"mangled pointer to %s in %s was poorly formed.\");\n",
			i,currentFunction->Name,data->ClassName);
		fprintf(fp,"      rb_raise(rb_eArgError,\"Invalid arguments (%s#%s)\");\n",data->ClassName,currentFunction->Name);
		fprintf(fp,"}\n"); 
		}
	      }

	    for (k = 0; k < (2 - (is_static || !is_vtkobject)); k++)
	      {
	      char methodname[256]; 
	      if (k == 0)
		{
		if (is_static)
		  {
		  fprintf(fp,"      {\n");
		  sprintf(methodname,"%s::%s",
			  data->ClassName,currentFunction->Name);
		  }
		else if (!is_vtkobject)
		  {
		  fprintf(fp,"      {\n");
		  sprintf(methodname,"op->%s",currentFunction->Name);
		  }
		else
		  {
		  if (currentFunction->IsPureVirtual)
		    {
		    fprintf(fp,"    if (strcmp(op->GetClassName(),\"%s\")==0)\n      {\n",data->ClassName);
		    fprintf(fp,"      rb_raise(rb_eTypeError,\"pure virtual method call\");\n");
		    fprintf(fp,"    }\n");
		    continue;
		    }
		  else
		    {
		    fprintf(fp,"    if (strcmp(op->GetClassName(),\"%s\")==0)\n      {\n",data->ClassName);
		    sprintf(methodname,"op->%s::%s",
			    data->ClassName,currentFunction->Name);
		    }
		  }
		}
	      else
		{
		fprintf(fp,"    else\n      {\n");
		sprintf(methodname,"op->%s",currentFunction->Name);
		}
                
	      switch (currentFunction->ReturnType % 0x1000)
		{
		case 0x2:
		  fprintf(fp,"      %s(",methodname);
		  break;
		case 0x109:
		  fprintf(fp,"      temp%i = &%s(",MAX_ARGS,methodname);
		  break;
		default:
		  fprintf(fp,"      temp%i = %s(",MAX_ARGS,methodname);
		}

	      for (i = 0; i < currentFunction->NumberOfArguments; i++)
		{
		if (i)
		  fprintf(fp,",");
		if (currentFunction->ArgTypes[i] % 0x1000 == 0x109)
		  fprintf(fp,"*(temp%i)",i);
		else if (currentFunction->NumberOfArguments == 1 
			 && currentFunction->ArgTypes[i] == 0x5000)
		  fprintf(fp,"((temp0 != Qnil) ? vtkRubyVoidFunc : NULL),(void *)temp%i",i);
		else
		  fprintf(fp,"temp%i",i);
		}
	      fprintf(fp,");\n");
          
	      if (currentFunction->NumberOfArguments == 1 
		  && currentFunction->ArgTypes[0] == 0x5000)
		fprintf(fp,"    %sArgDelete(vtkRubyVoidFuncArgDelete);\n",
			methodname);
	      fprintf(fp,"      }\n");
	      }

	    for (i = 0; i < currentFunction->NumberOfArguments; i++)
	      {
	      if (currentFunction->ArgCounts[i] &&  /* array */
		  currentFunction->ArgTypes[i] % 0x10 != 0 && /* not a special type */
		  currentFunction->ArgTypes[i] % 0x10 != 0x9 && /* not class pointer */
		  currentFunction->ArgTypes[i] % 0x10 != 0x8 && 
		  currentFunction->ArgTypes[i] % 0x10 != 0x2 && /* not void pointer */
		  (currentFunction->ArgTypes[i] % 0x2000 < 0x1000)) /* not const */
		{
		fprintf(fp,"    if (vtkRubyCheckArray(argv,%d,temp%d,%d))\n      {\n"
			"      return 0;\n"
			"      }\n", i, i, currentFunction->ArgCounts[i]);
		}
	      }
	    do_return(fp);
	    fprintf(fp,"    }\n");
	    if(currentFunction->NumberOfArguments != 0)
	      fprintf(fp,"  break%d:\n",occ);
	    }
	  }
	}
      fprintf(fp,"  rb_raise(rb_eArgError,\"wrong # of arguments (%%d)\",argc);\n");
      fprintf(fp,"  return Qnil;\n");
      fprintf(fp,"}\n");
      if(currentFunction->IsLegacy)
        {
        fprintf(fp,"#endif\n");
        }
      fprintf(fp,"\n");

      /* clear all occurances of this method from further consideration */
      for (occ = fnum + 1; occ < numberOfWrappedFunctions; occ++)
        {
        /* is it the same name */
        if (wrappedFunctions[occ]->Name && 
            !strcmp(theFunc->Name,wrappedFunctions[occ]->Name))
          {
          int siglen = (int)strlen(wrappedFunctions[fnum]->Signature);
          /* memory leak here but ... */
          wrappedFunctions[occ]->Name = NULL;
          wrappedFunctions[fnum]->Signature = (char *)
            realloc(wrappedFunctions[fnum]->Signature,siglen+3+
                    strlen(wrappedFunctions[occ]->Signature));
          strcpy(&wrappedFunctions[fnum]->Signature[siglen],"\\n");
          strcpy(&wrappedFunctions[fnum]->Signature[siglen+2],
                 wrappedFunctions[occ]->Signature);
          }
        }
      } /* is this method non NULL */
    } /* loop over all methods */
  
}



void outputFunction(FILE *fp, FileInfo *data)
{
  int i;
  int args_ok = 1;

  fp = fp;
  /* some functions will not get wrapped no matter what else,
     and some really common functions will appear only in vtkObjectRuby */
  if (currentFunction->IsOperator || 
      currentFunction->ArrayFailure ||
      !currentFunction->IsPublic ||
      !currentFunction->Name)
    {
    return;
    }
  
  /* check to see if we can handle the args */
  for (i = 0; i < currentFunction->NumberOfArguments; i++)
    {
    if (currentFunction->ArgTypes[i] % 0x1000 == 9) args_ok = 0;
    if ((currentFunction->ArgTypes[i] % 0x10) == 8) args_ok = 0;
    if (((currentFunction->ArgTypes[i] % 0x1000)/0x100 != 0x3)&&
        (currentFunction->ArgTypes[i] % 0x1000 != 0x109)&&
        ((currentFunction->ArgTypes[i] % 0x1000)/0x100)) args_ok = 0;
    if (currentFunction->ArgTypes[i] % 0x1000 == 0x313) args_ok = 0;
    if (currentFunction->ArgTypes[i] % 0x1000 == 0x314) args_ok = 0;
    if (currentFunction->ArgTypes[i] % 0x1000 == 0x31A) args_ok = 0;
    if (currentFunction->ArgTypes[i] % 0x1000 == 0x31B) args_ok = 0;
    if (currentFunction->ArgTypes[i] % 0x1000 == 0x31C) args_ok = 0;
    if (currentFunction->ArgTypes[i] % 0x1000 == 0x315) args_ok = 0;
    if (currentFunction->ArgTypes[i] % 0x1000 == 0x316) args_ok = 0;
    }
  if ((currentFunction->ReturnType % 0x10) == 0x8) args_ok = 0;
  if (currentFunction->ReturnType % 0x1000 == 0x9) args_ok = 0;
  if (((currentFunction->ReturnType % 0x1000)/0x100 != 0x3)&&
      (currentFunction->ReturnType % 0x1000 != 0x109)&&
      ((currentFunction->ReturnType % 0x1000)/0x100)) args_ok = 0;


  /* eliminate unsigned char * and unsigned short * */
  if (currentFunction->ReturnType % 0x1000 == 0x313) args_ok = 0;
  if (currentFunction->ReturnType % 0x1000 == 0x314) args_ok = 0;
  if (currentFunction->ReturnType % 0x1000 == 0x31A) args_ok = 0;
  if (currentFunction->ReturnType % 0x1000 == 0x31B) args_ok = 0;
  if (currentFunction->ReturnType % 0x1000 == 0x31C) args_ok = 0;
  if (currentFunction->ReturnType % 0x1000 == 0x315) args_ok = 0;
  if (currentFunction->ReturnType % 0x1000 == 0x316) args_ok = 0;

  if (currentFunction->NumberOfArguments && 
      (currentFunction->ArgTypes[0] == 0x5000)
      &&(currentFunction->NumberOfArguments != 0x1)) args_ok = 0;

  /* make sure we have all the info we need for array arguments in */
  for (i = 0; i < currentFunction->NumberOfArguments; i++)
    {
    if (((currentFunction->ArgTypes[i] % 0x1000)/0x100 == 0x3)&&
        (currentFunction->ArgCounts[i] <= 0)&&
        (currentFunction->ArgTypes[i] % 0x1000 != 0x309)&&
        (currentFunction->ArgTypes[i] % 0x1000 != 0x303)&&
        (currentFunction->ArgTypes[i] % 0x1000 != 0x302)) args_ok = 0;
    }

  /* if we need a return type hint make sure we have one */
  switch (currentFunction->ReturnType % 0x1000)
    {
    case 0x301: case 0x307: case 0x30A: case 0x30B: case 0x30C: case 0x30D:
    case 0x304: case 0x305: case 0x306:
      args_ok = currentFunction->HaveHint;
      break;
    }
  
  /* make sure it isn't a Delete or New function */
  if (!strcmp("Delete",currentFunction->Name) ||
      !strcmp("New",currentFunction->Name))
    {
    args_ok = 0;
    }
  
  /* check for New() function */
  if (!strcmp("New",currentFunction->Name) &&
      currentFunction->NumberOfArguments == 0)
    {
    class_has_new = 1;
    }

  if (currentFunction->IsPublic && args_ok && 
      strcmp(data->ClassName,currentFunction->Name) &&
      strcmp(data->ClassName, currentFunction->Name + 1))
    {
    wrappedFunctions[numberOfWrappedFunctions] = currentFunction;
    numberOfWrappedFunctions++;
    }
}


void outputMethods(FILE *fp, FileInfo *data)
{
  int fnum;

  for (fnum = 0; fnum < numberOfWrappedFunctions; fnum++)
    {
    if(wrappedFunctions[fnum]->IsLegacy)
      {
      fprintf(fp,"#if !defined(VTK_LEGACY_REMOVE)\n");
      }
    if(wrappedFunctions[fnum]->Name)
      {
      if (((wrappedFunctions[fnum]->ReturnType/0x1000) & 0x2) == 0x2)
        fprintf(fp,"  rb_define_singleton_method(c%s, \"%s\", VALUEFUNC(rb_%s_%s), -1);\n",
		data->ClassName, wrappedFunctions[fnum]->Name,
		data->ClassName, wrappedFunctions[fnum]->Name);
      else
	fprintf(fp,"  rb_define_method(c%s, \"%s\", VALUEFUNC(rb_%s_%s), -1);\n",
		data->ClassName, wrappedFunctions[fnum]->Name,
		data->ClassName, wrappedFunctions[fnum]->Name);
      }
    if(wrappedFunctions[fnum]->IsLegacy)
      {
      fprintf(fp,"#endif\n");
      }
    }

  if (strcmp("vtkObject",data->ClassName)==0)
    {
    fprintf(fp,"  rb_define_method(c%s, \"AddObserver\", VALUEFUNC(rb_%s_AddObserver), -1);\n",
	    data->ClassName, data->ClassName);
    }
  else if (!strcmp("vtkObjectBase",data->ClassName))
    {
    fprintf(fp,"  rb_define_method(c%s, \"GetAddressAsString\", VALUEFUNC(rb_%s_GetAddressAsString), -1);\n",
	    data->ClassName, data->ClassName);
    fprintf(fp,"  rb_define_method(c%s, \"PrintRevisions\", VALUEFUNC(rb_%s_PrintRevisions), 0);\n",
	    data->ClassName, data->ClassName);
    }

}


/* print the parsed structures */
void vtkParseOutput(FILE *fp, FileInfo *data)
{
  int i;
  
  fprintf(fp,"// ruby wrapper for %s object\n//\n",data->ClassName);
  fprintf(fp,"#define VTK_WRAPPING_CXX\n");
  if (strcmp("vtkObjectBase",data->ClassName) != 0)
    {
    /* Block inclusion of full streams.  */
    fprintf(fp,"#define VTK_STREAMS_FWD_ONLY\n");
    }
  #if !defined(__APPLE__)
  fprintf(fp,"#include \"vtkRuby.h\"\n");
  fprintf(fp,"#undef _XOPEN_SOURCE /* Conflicts with standards.h.  */\n");
  fprintf(fp,"#undef _THREAD_SAFE /* Conflicts with pthread.h.  */\n");
  fprintf(fp,"#include \"vtkRubyUtil.h\"\n");
  #endif
  fprintf(fp,"#include \"%s.h\"\n",data->ClassName);
  #if defined(__APPLE__)
  fprintf(fp,"#include \"vtkRubyUtil.h\"\n\n");
  #endif

  fprintf(fp,"#ifdef __cplusplus\n");
  fprintf(fp,"#  ifndef RUBY_METHOD_FUNC /* These definitions should work for Ruby 1.4.6 */\n");
  fprintf(fp,"#    define PROTECTFUNC(f) ((VALUE (*)()) f)\n");
  fprintf(fp,"#    define VALUEFUNC(f) ((VALUE (*)()) f)\n");
  fprintf(fp,"#    define VOIDFUNC(f)  ((void (*)()) f)\n");
  fprintf(fp,"#  else\n");
  fprintf(fp,"#    ifndef ANYARGS /* These definitions should work for Ruby 1.6 */\n");
  fprintf(fp,"#      define PROTECTFUNC(f) ((VALUE (*)()) f)\n");
  fprintf(fp,"#      define VALUEFUNC(f) ((VALUE (*)()) f)\n");
  fprintf(fp,"#      define VOIDFUNC(f)  ((RUBY_DATA_FUNC) f)\n");
  fprintf(fp,"#    else /* These definitions should work for Ruby 1.7+ */\n");
  fprintf(fp,"#      define PROTECTFUNC(f) ((VALUE (*)(VALUE)) f)\n");
  fprintf(fp,"#      define VALUEFUNC(f) ((VALUE (*)(ANYARGS)) f)\n");
  fprintf(fp,"#      define VOIDFUNC(f)  ((RUBY_DATA_FUNC) f)\n");
  fprintf(fp,"#    endif\n");
  fprintf(fp,"#  endif\n");
  fprintf(fp,"#else\n");
  fprintf(fp,"#  define VALUEFUNC(f) (f)\n");
  fprintf(fp,"#  define VOIDFUNC(f) (f)\n");
  fprintf(fp,"#endif\n\n");

  fprintf(fp,"#if defined(WIN32)\n");
  fprintf(fp,"extern \"C\" { __declspec( dllexport ) void Init_%s(VALUE); }\n",
          data->ClassName);
  fprintf(fp,"#else\n");
  fprintf(fp,"extern \"C\" { void Init_%s(VALUE); }\n",
          data->ClassName);
  fprintf(fp,"#endif\n\n");
  for (i = 0; i < data->NumberOfSuperClasses; i++)
    {
    fprintf(fp,"extern \"C\" { VALUE Init_%s(VALUE); }\n",
            data->SuperClasses[i]);
    }
  
  if (!strcmp("vtkObject",data->ClassName))
    {
    /* Add the AddObserver method to vtkObject. */
    fprintf(fp,"static VALUE rb_vtkObject_AddObserver(int argc, VALUE *argv, VALUE self)\n");
    fprintf(fp,"{\n");
    fprintf(fp,"  vtkObject *op;\n");
    fprintf(fp,"  char *temp0;\n");
    fprintf(fp,"  VALUE temp1;\n");
    fprintf(fp,"  float temp2;\n");
    fprintf(fp,"  unsigned long     temp20 = 0;\n");
    fprintf(fp,"  if (argc!=2&&argc!=3)\n");
    fprintf(fp,"    rb_raise(rb_eArgError,\"wrong # of arguments (%%d for 2)\",argc);\n");
    fprintf(fp,"  op = (vtkObject *)vtkRubyGetPointerFromObject(self);\n");
    fprintf(fp,"  temp0 = STR2CSTR(argv[0]);\n");
    fprintf(fp,"  if(rb_class_of(argv[1])!=rb_cProc)\n");
    fprintf(fp,"    rb_raise(rb_eArgError,\"the second argument must be Proc\");\n");
    fprintf(fp,"  temp1 = argv[1];\n");
    fprintf(fp,"  vtkRubyCommand *cbc = vtkRubyCommand::New();\n");
    fprintf(fp,"  cbc->SetObject(temp1);\n");
    fprintf(fp,"  if (argc==2)\n");
    fprintf(fp,"    {\n");
    fprintf(fp,"    temp20 = op->AddObserver(temp0,cbc);\n");
    fprintf(fp,"    }\n");
    fprintf(fp,"  else if(argc==3)\n");
    fprintf(fp,"    {\n");
    fprintf(fp,"    temp2 = NUM2DBL(argv[2]);\n");
    fprintf(fp,"    temp20 = op->AddObserver(temp0,cbc,temp2);\n");
    fprintf(fp,"    }\n");
    fprintf(fp,"  cbc->Delete();\n");
    fprintf(fp,"  return ULONG2NUM(temp20);\n");
    fprintf(fp,"}\n\n");
    }

  if (!strcmp("vtkObjectBase",data->ClassName))
    {
    /* while we are at it spit out the GetStringFromObject method */
    fprintf(fp,"VALUE rb_vtkObjectBase_GetAddressAsString(int argc, VALUE *argv, VALUE self)\n");
    fprintf(fp,"{\n");

    /* declare the variables */
    fprintf(fp,"  %s *op;\n",data->ClassName);

    /* handle unbound method call if 'self' is a RuVTKClass */
    fprintf(fp,"  char *typecast;\n\n");
    fprintf(fp,"  char temp20[256];\n");
    fprintf(fp,"  if(argc==1)\n");
    fprintf(fp,"    typecast = STR2CSTR(argv[0]);\n");
    fprintf(fp,"  else if(argc>1)\n");
    fprintf(fp,"    rb_raise(rb_eArgError,\"wrong # of arguments (%%d for <=1)\", argc);\n");
    fprintf(fp,"  op = (%s *)vtkRubyGetPointerFromObject(self);\n",data->ClassName);
    fprintf(fp,"  sprintf(temp20,\"Addr=%%p\",op);\n");
    fprintf(fp,"  return rb_str_new2(temp20);\n");
    fprintf(fp,"}\n\n");

    /* Add the PrintRevisions method to vtkObjectBase. */
    fprintf(fp,"VALUE rb_vtkObjectBase_PrintRevisions(int argc, VALUE *args, VALUE self)\n");
    fprintf(fp,"{\n");
    fprintf(fp,"  %s *op;\n",data->ClassName);
    fprintf(fp,"  op = (%s *)vtkRubyGetPointerFromObject(self);\n",data->ClassName);
    fprintf(fp,"  ostrstream vtkmsg_with_warning_C4701;\n");
    fprintf(fp,"  op->PrintRevisions(vtkmsg_with_warning_C4701);\n");
    fprintf(fp,"  vtkmsg_with_warning_C4701.put('\\0');\n");
    fprintf(fp,"  VALUE result = rb_str_new2(vtkmsg_with_warning_C4701.str());\n");
    fprintf(fp,"  delete vtkmsg_with_warning_C4701.str();\n");
    fprintf(fp,"  return result;\n");
    fprintf(fp,"}\n\n");
    }
  
  /* insert function handling code here */
  for (i = 0; i < data->NumberOfFunctions; i++)
    {
    currentFunction = data->Functions + i;
    outputFunction(fp, data);
    }
  if (data->NumberOfSuperClasses || !data->IsAbstract)
    {
    outputFunction2(fp, data);
    }
  
  /* output the class initilization function */
  if (strcmp(data->ClassName,"vtkObjectBase") == 0)
    { /* special wrapping for vtkObject */
    if (class_has_new)
      {
      fprintf(fp,"static VALUE rb_%sNew(int argc, VALUE *argv, VALUE klass)\n",data->ClassName);
      fprintf(fp,"{\n");
      fprintf(fp,"  if (argc==0)\n");
      fprintf(fp,"    return rb_VTKObject_New(klass, (vtkObjectBase*)%s::New());\n",
	      data->ClassName);
      fprintf(fp,"  else if (argc==1)\n");
      fprintf(fp,"    return vtkRubyGetObjectFromObject(argv[0], \"%s\", klass);\n",
	      data->ClassName);
      fprintf(fp,"  else\n");
      fprintf(fp,"    rb_raise(rb_eArgError,\"wrong # of arguments (%%d for 0 or 1)\",argc);\n");
      fprintf(fp,"\n}\n\n");
      }
    fprintf(fp,"static VALUE c%s;\n", data->ClassName);
    fprintf(fp,"static void rb_VTKClass_%sNew(VALUE module)\n", data->ClassName);
    fprintf(fp,"{\n");
    fprintf(fp,"  c%s = rb_vtk_define_class_under(module, \"%s\", rb_cObject);\n",
	    data->ClassName, data->ClassName);
    if (class_has_new)
      {
      fprintf(fp,"  rb_define_singleton_method(c%s, \"new\", VALUEFUNC(rb_%sNew), -1);\n",
	      data->ClassName, data->ClassName);
      }
    outputMethods(fp, data);
    fprintf(fp,"}\n\n");
    fprintf(fp,"static int c%s_flag = 1;\n", data->ClassName);
    fprintf(fp,"#if defined(WIN32)\n");
    fprintf(fp,"extern __declspec( dllexport )\n");
    fprintf(fp,"#else\n");
    fprintf(fp,"extern \n");
    fprintf(fp,"#endif\n");
    fprintf(fp,"VALUE rb_%sClass(VALUE module)\n", data->ClassName);
    fprintf(fp,"{\n");
    fprintf(fp,"  if (c%s_flag)\n", data->ClassName);
    fprintf(fp,"    {\n");
    fprintf(fp,"    rb_VTKClass_%sNew(module);\n", data->ClassName);
    fprintf(fp,"    c%s_flag = 0;\n", data->ClassName);
    fprintf(fp,"    }\n");
    fprintf(fp,"  return c%s;\n", data->ClassName);
    fprintf(fp,"}\n");
    fprintf(fp,"#if defined(WIN32)\n");
    fprintf(fp,"extern __declspec( dllexport )\n");
    fprintf(fp,"#else\n");
    fprintf(fp,"extern \n");
    fprintf(fp,"#endif\n");
    fprintf(fp,"VALUE rb_%sClass()\n", data->ClassName);
    fprintf(fp,"{\n");
    fprintf(fp,"  if (c%s_flag)\n", data->ClassName);
    fprintf(fp,"    {\n");
    fprintf(fp,"    rb_raise(rb_eRuntimeError,\"%sClass is not defined\");\n", data->ClassName);
    fprintf(fp,"    }\n");
    fprintf(fp,"  return c%s;\n", data->ClassName);
    fprintf(fp,"}\n\n");
    fprintf(fp,"#ifdef __cplusplus\n");
    fprintf(fp,"extern \"C\" {\n");
    fprintf(fp,"#endif\n");
    fprintf(fp,"void Init_%s(VALUE module)\n{\n",data->ClassName);
    fprintf(fp,"  rb_%sClass(module);\n", data->ClassName);
    fprintf(fp,"}\n");
    fprintf(fp,"#ifdef __cplusplus\n");
    fprintf(fp,"}\n");
    fprintf(fp,"#endif\n");
    }
  else if (data->NumberOfSuperClasses)
    { /* wrapping of descendants of vtkObjectBase */
    if (class_has_new)
      {
      fprintf(fp,"static VALUE rb_%sNew(int argc, VALUE *argv, VALUE klass)\n",data->ClassName);
      fprintf(fp,"{\n");
      fprintf(fp,"  if (argc==0)\n");
      fprintf(fp,"    return rb_VTKObject_New(klass, (vtkObjectBase*)%s::New());\n",
	      data->ClassName);
      fprintf(fp,"  else if (argc==1)\n");
      fprintf(fp,"    return vtkRubyGetObjectFromObject(argv[0], \"%s\", klass);\n",
	      data->ClassName);
      fprintf(fp,"  else\n");
      fprintf(fp,"    rb_raise(rb_eArgError,\"wrong # of arguments (%%d for 0 or 1)\",argc);\n");
      fprintf(fp,"\n}\n\n");
      }
    fprintf(fp,"static VALUE c%s;\n", data->ClassName);
    fprintf(fp,"static void rb_VTKClass_%sNew(VALUE module)\n", data->ClassName);
    fprintf(fp,"{\n");
    fprintf(fp,"#if defined(WIN32)\n");
    fprintf(fp,"  extern __declspec( dllimport )\n");
    fprintf(fp,"#else\n");
    fprintf(fp,"  extern \n");
    fprintf(fp,"#endif\n");
    fprintf(fp,"  VALUE rb_%sClass(VALUE);\n", data->SuperClasses[0]);
    fprintf(fp,"  c%s = rb_vtk_define_class_under(module, \"%s\", rb_%sClass(module));\n",
	    data->ClassName, data->ClassName, data->SuperClasses[0]);
    if (class_has_new)
      {
      fprintf(fp,"  rb_define_singleton_method(c%s, \"new\", VALUEFUNC(rb_%sNew), -1);\n",
	      data->ClassName, data->ClassName);
      }
    outputMethods(fp, data);
    fprintf(fp,"}\n");
    fprintf(fp,"static int c%s_flag = 1;\n", data->ClassName);
    fprintf(fp,"#if defined(WIN32)\n");
    fprintf(fp,"extern __declspec( dllexport )\n");
    fprintf(fp,"#else\n");
    fprintf(fp,"extern \n");
    fprintf(fp,"#endif\n");
    fprintf(fp,"VALUE rb_%sClass(VALUE module)\n", data->ClassName);
    fprintf(fp,"{\n");
    fprintf(fp,"  if (c%s_flag)\n", data->ClassName);
    fprintf(fp,"    {\n");
    fprintf(fp,"    rb_VTKClass_%sNew(module);\n", data->ClassName);
    fprintf(fp,"    c%s_flag = 0;\n", data->ClassName);
    fprintf(fp,"    }\n");
    fprintf(fp,"  return c%s;\n", data->ClassName);
    fprintf(fp,"}\n");
    fprintf(fp,"#ifdef __cplusplus\n");
    fprintf(fp,"extern \"C\" {\n");
    fprintf(fp,"#endif\n");
    fprintf(fp,"void Init_%s(VALUE module)\n{\n",data->ClassName);
    fprintf(fp,"  rb_%sClass(module);\n", data->ClassName);
    fprintf(fp,"}\n");
    fprintf(fp,"#ifdef __cplusplus\n");
    fprintf(fp,"}\n");
    fprintf(fp,"#endif\n");
    }
  else if (!data->IsAbstract)
    { /* wrapping of 'special' non-vtkObject classes */
    fprintf(fp,"static VALUE rb_%sNew(int argc, VALUE *argv, VALUE klass)\n",
	    data->ClassName);
    fprintf(fp,"{\n");
    fprintf(fp,"  VALUE %s *obj = new %s;\n",data->ClassName,data->ClassName);
    fprintf(fp,"  return rb_VTKSpecialObject_New(klass, (void*)obj);\n");
    fprintf(fp,"\n}\n\n");
    fprintf(fp,"static VALUE c%s;\n", data->ClassName);
    fprintf(fp,"static void rb_VTKClass_%sNew(VALUE module)\n", data->ClassName);
    fprintf(fp,"{\n");
    fprintf(fp,"  c%s = rb_vtk_define_class_under(module, \"%s\", rb_cObject);\n",
	    data->ClassName, data->ClassName);
    fprintf(fp,"  rb_define_singleton_method(c%s, \"new\", VALUEFUNC(rb_%sNew), -1);\n",
	      data->ClassName, data->ClassName);
    fprintf(fp,"}\n");
    fprintf(fp,"static int c%s_flag = 1;\n", data->ClassName);
    fprintf(fp,"#if defined(WIN32)\n");
    fprintf(fp,"extern __declspec( dllexport )\n");
    fprintf(fp,"#else\n");
    fprintf(fp,"extern \n");
    fprintf(fp,"#endif\n");
    fprintf(fp,"VALUE rb_%sClass(VALUE module)\n", data->ClassName);
    fprintf(fp,"{\n");
    fprintf(fp,"  if (c%s_flag)\n", data->ClassName);
    fprintf(fp,"    {\n");
    fprintf(fp,"    rb_VTKClass_%sNew(module);\n", data->ClassName);
    fprintf(fp,"    c%s_flag = 0;\n", data->ClassName);
    fprintf(fp,"    }\n");
    fprintf(fp,"  return c%s;\n", data->ClassName);
    fprintf(fp,"}\n");
    fprintf(fp,"#ifdef __cplusplus\n");
    fprintf(fp,"extern \"C\" {\n");
    fprintf(fp,"#endif\n");
    fprintf(fp,"void Init_%s(VALUE module)\n{\n",data->ClassName);
    fprintf(fp,"  rb_%sClass(module);\n", data->ClassName);
    fprintf(fp,"}\n");
    fprintf(fp,"#ifdef __cplusplus\n");
    fprintf(fp,"}\n");
    fprintf(fp,"#endif\n");
    }
  else
    { /* un-wrappable classes */
    fprintf(fp,"#ifdef __cplusplus\n");
    fprintf(fp,"extern \"C\" {\n");
    fprintf(fp,"#endif\n");
    fprintf(fp,"void Init_%s(VALUE module)\n{\n",data->ClassName);
    fprintf(fp,"  rb_raise(rb_eRuntimeError, \"%s Class is not wrappable\");\n");
    fprintf(fp,"}\n");
    fprintf(fp,"#ifdef __cplusplus\n");
    fprintf(fp,"}\n");
    fprintf(fp,"#endif\n");
    }

}

